/*
 * Gauss-Seidel Descendant PageRank Implementation
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "sparse_matrix.h"
#include "vector.h"


double* gauss_seidel_pagerank(const SparseMatrix *matrix, double epsilon, double alpha) {
    int N = matrix->num_nodes;
    size_t vec_size = N * sizeof(double);

    double *x = malloc(vec_size);
    double *x_old = malloc(vec_size);
    double *f = malloc(vec_size);
    double right_const = (1.0 - alpha) / N;

    if (!x || !x_old || !f) {
        fprintf(stderr, "Memory allocation failed\n");
        exit(EXIT_FAILURE);
    }

    init_vector(x, N, 1.0 / N);
    generate_f(matrix, f);

    double diff;
    int iter = 0;

    do {
        // 1. store x from previous step
        memcpy(x_old, x, vec_size);

        // 2. alpha/N * (x * f)
        double x_f = vec_product(x_old, f, N);
        double right_var = (alpha / (double)N) * x_f;

        // 3. alpha*(pi*M) + (right_const+alpha/N * (pi * f))*e
        //    splitted in 2 parts (2 sums instead of one)
        //    Can't use previously defined multiply_vector_matrix_parallel function here
        //
        //    sum from 1 to i-1: use old x
        //    i=j: use value M(i,i)
        //    sum from i+1 to n: use x
        //
        //    Gauss Seidel descending, so i from n to 0
        //
        for (int i = N - 1; i >= 0; i--) {
            double result_mult = 0.0;
            double M_ii = 0.0;
            for (int k = matrix->row_ptr[i]; k < matrix->row_ptr[i + 1]; k++) {
                int j = matrix->arcs[k].dest;
                double value = matrix->arcs[k].value;

                if (j == i) {
                    M_ii = value;
                } else if (j > i) {
                    result_mult += value * x[j];
                } else {
                    result_mult += value * x_old[j];
                }
            }
            double diag_dependant_term = 1.0 - alpha * M_ii;

            x[i] = (alpha * result_mult + right_const + right_var) / diag_dependant_term;
        }

        // 4. renormalization step
        normalize_vector(x, N);

        // 5. convergence
        diff = diff_norm_vector(x, x_old, N);

        // if ((++iter) % 10 == 0) {
        //     printf("Gauss-Seidel Iteration %d: diff = %.16f\n", iter, diff);
        // }
        ++iter;
    } while (diff > epsilon);

    printf("Gauss-Seidel: %d Iterations\n", iter);

    free(x_old);
    free(f);

    // x ~ pi (convergence)
    return x;
}
