#include <stdio.h>
#include <stdlib.h>
#include "matrix_operation.h"
#include "sparse_matrix.h"
#include "read_from_mtx.h"
#include "read_from_rb.h"
#include "vector.h"
#include "power_algorithm.h"
#include "gauss_seidel.h"
#include "time_helper.h"
#include <sys/time.h>
#include <string.h>

void test_stationary_distribution(const char *path, double epsilon, double alpha) {
    struct timeval tvstart, tvend;

    gettimeofday(&tvstart, NULL);

    SparseMatrix *matrix = NULL;
    const char *extension = strrchr(path, '.');
    if (extension && strcasecmp(extension, ".rb") == 0) {
        matrix = read_sparse_matrix_from_rb(path);
    } else {
        matrix = read_sparse_matrix_from_mtx(path);
    }

    convert_to_stochastic(matrix);
    gettimeofday(&tvend, NULL);
    print_time_diff("Read and convert matrix", &tvstart, &tvend);

    gettimeofday(&tvstart, NULL);
    double *pi_power = pagerank(matrix, epsilon, alpha);
    gettimeofday(&tvend, NULL);
    print_time_diff("Pagerank", &tvstart, &tvend);

    gettimeofday(&tvstart, NULL);
    double *pi_new = gauss_seidel_pagerank(matrix, epsilon, alpha);

    gettimeofday(&tvend, NULL);
    print_time_diff("Gauss-Seidel", &tvstart, &tvend);

    double diff = diff_norm_vector(pi_power, pi_new, matrix->num_nodes);
    printf("Difference norm Pagerank and Gauss-Seidel: %.16f\n", diff);

    free(pi_power);
    free(pi_new);
    free_sparse_matrix(matrix);
}

int main(int argc, char *argv[]) {
    const char *matrix_path = NULL;
    double epsilon = -1.0;
    double alpha = -1.0;

    for (int i = 1; i < argc; i++) {
        if ((strcmp(argv[i], "--matrix") == 0 || strcmp(argv[i], "-matrix") == 0) && i + 1 < argc) {
            matrix_path = argv[++i];
        } else if ((strcmp(argv[i], "--epsilon") == 0 || strcmp(argv[i], "-epsilon") == 0) && i + 1 < argc) {
            epsilon = atof(argv[++i]);
        } else if ((strcmp(argv[i], "--alpha") == 0 || strcmp(argv[i], "-alpha") == 0) && i + 1 < argc) {
            alpha = atof(argv[++i]);
        } else {
            fprintf(stderr, "Unknown or incomplete argument: %s\n", argv[i]);
            return EXIT_FAILURE;
        }
    }

    if (matrix_path == NULL || epsilon <= 0 || alpha < 0 || alpha > 1) {
        fprintf(stderr, "Usage: %s --matrix <path> --epsilon <value> --alpha <value>\n", argv[0]);
        return EXIT_FAILURE;
    }

    test_stationary_distribution(matrix_path, epsilon, alpha);
    return 0;
}
